#include <Wire.h>
#include <Ticker.h>
#include <ESP8266WiFi.h>
#include <cmath>

// <-------------------- I2C  funkciji zacetek -------------------------->
void I2CWriteRegister(uint8_t I2CDevice, uint8_t RegAdress, uint8_t Value){
  // I2CDevice - Naslov I2C naprave
  // RegAddress - Naslov registra 
  // Value - Vrednost za vpisati v register
  
  Wire.beginTransmission(I2CDevice);
  // Napravi sporočimo naslov registra, s katerega želimo brati:
  Wire.write(RegAdress);
  // Posljemo vrednost
  Wire.write(Value);
  Wire.endTransmission();
}

void I2CReadRegister(uint8_t I2CDevice, uint8_t RegAdress, uint8_t NBytes, uint8_t *Value){
  Wire.beginTransmission(I2CDevice);
  // Napravi sporočimo naslov registra, s katerega želimo brati:
  Wire.write(RegAdress);
  // Končamo prenos:
  Wire.endTransmission();
  
  // Napravi sporočimo, da želimo prebrati določeno število 8-bitnih registrov:
  Wire.requestFrom(I2CDevice, NBytes);
  for (int q = 0; q < NBytes; q++) {
    // Preberemo naslednji 8-bitni register oz. naslednji bajt:
    *Value = (uint8_t)Wire.read();
    Value++;
    //uint32_t vrednost = Wire.read();
  }

}

// <-------------------- I2C  Funkcije konec -------------------------->

// Naslov MPU9250 na I2C vodilu
#define MPU_ADD 104
// Naslov registra za pospesek 
#define ACC_MEAS_REG 59
// Naslov registra za ziroskop
#define GYRO_MEAS_REG 67

// Zastavice za omogočanje izpisa
#define PRINT_CORR 0
#define PRINT_RAW 0

// Časovnik za branje senzorja
Ticker readSensor;

// Konstante uporabljene za meritve
#define MEAS_AVG 4        // Število povprečenih vzorcev
#define MEAS_BUF 100      // Število hranjenih meritev
#define MEAS_PERIOD 0.05  // Frekvenca vzorčenja

float measurements[MEAS_BUF]; // Meritve (ciklični buffer)
int buf_index = 0;            // Kazalec na naslednji prosti prostor za meritev

// Podatki za dostopno točko WiFi
#define NAZIV "Električni štedilnik"
#define GESLO "passwort"

// Podatki za povezavo na ThingSpeak
#define CHANNEL 2145640
#define TS_WRITE_API "O13XS7WCM8H9C5HD"

WiFiClient client;

void MPU9250_init(){
  // Resetiraj MPU9250 senzora => Register PWR_MGMT_1 (107)
  I2CWriteRegister(MPU_ADD,107,128); // 128 = 1000 0000
  // Pocakaj
  delay(500);
  // Preveri ID od senzora => Register WHO_AM_I (117) 
  uint8_t ID;
  I2CReadRegister(MPU_ADD,117,1,&ID);
  Serial.println("ID:");
  Serial.println(ID, HEX);
  // Gyroscope Conf => Register GYRO_CONFIG (27) 
  // 4 in 3 bit dolocata obseg 
  I2CWriteRegister(MPU_ADD,27,0); // 
  delay(100);
  // Accelerator Conf => Register ACCEL_CONFIG (28)
  // 4 in 3 bit dolocata obseg 
  // Opciono => Register ACCEL_CONFIG_2 (29)
  I2CWriteRegister(MPU_ADD,28,0); // 
  delay(100);
}

float accStep = 9.81 / (1 << 14); // Konstanta za pretvorbo senzorskih podatkov

// Ciklični bufferji za povprečenje vzorcev
float avgX_buf[MEAS_AVG];
float avgY_buf[MEAS_AVG];
float avgZ_buf[MEAS_AVG];
int avg_index = 0; // Indeks bufferjev za povprečenje

// Prostor za prebrane vzorce
uint8_t accMeas[] = {0,0,0,0,0,0};

void readAcc(){
  // Prenos podatkov iz senzorja na mikrokrmilnik
  I2CReadRegister(MPU_ADD,ACC_MEAS_REG,6,(uint8_t*) &accMeas);

  // Izračun x, y in z vrednosti
  int32_t tmp = ((int8_t)accMeas[0]<<8) + ((uint8_t)accMeas[1]);
  avgX_buf[avg_index] = tmp*accStep;
  tmp = ((int8_t)accMeas[2]<<8) + ((uint8_t)accMeas[3]);
  avgY_buf[avg_index] = tmp*accStep;
  tmp = ((int8_t)accMeas[4]<<8) + ((uint8_t)accMeas[5]);
  avgZ_buf[avg_index] = tmp*accStep;
  avg_index = (avg_index + 1)%4;

  // Povprečenje zadnjih 4 vzorcev (glajenje)
  float ax = (avgX_buf[0] + avgX_buf[1] + avgX_buf[2] + avgX_buf[3])/4;
  float ay = (avgY_buf[0] + avgY_buf[1] + avgY_buf[2] + avgY_buf[3])/4;
  float az = (avgZ_buf[0] + avgZ_buf[1] + avgZ_buf[2] + avgZ_buf[3])/4;

  // Izračun magnitude vektorja pospeška
  float magn = sqrt(ax*ax + ay*ay + az*az);

#if PRINT_RAW
  Serial.printf("%.2f,%.2f,%.2f,%.2f\n", ax, ay, az, magn);
#endif
  
  // Zapis meritve
  measurements[buf_index] = magn;
  buf_index = (buf_index + 1) % MEAS_BUF;
}

// Širina okna za korelacijo
#define N_SAMP ((int)(2/MEAS_PERIOD))
float correlations[N_SAMP];

// Funkcija za izračun korelacije (in korakov)
float calculateSteps() {
  int buf_ind = buf_index;                    // Shrani trenutno vrednost indeksa
  int startj = buf_ind - N_SAMP + MEAS_BUF;   // Začetni indeks predloge ('template')

#if PRINT_CORR
  Serial.printf("Correllations:\n");
#endif
  
  // Izračun povprečne magnitude v celotnem oknu
  float avg = 0.0;
  for(int i = 0; i < 2*N_SAMP; i++)
  {
    int index = (buf_ind - 2*N_SAMP + i + MEAS_BUF) % MEAS_BUF;
    avg += measurements[index];
  }
  avg /= 2*N_SAMP;

  // Izračun korelacije
  // Premikanje po vseh odmikih
  for(int i = 0; i < N_SAMP; i++)
  {
    // Izračun korelacije na odmiku i
    correlations[i] = 0;
    int starti = buf_ind - 2*N_SAMP + i + MEAS_BUF;
    for(int j = 0; j < N_SAMP; j++)
    {
      correlations[i] += (measurements[(starti + j)%MEAS_BUF]-avg)*(measurements[(startj + j)%MEAS_BUF]-avg);
    }

#if PRINT_CORR
    Serial.printf("%.2f,", correlations[i]);
#endif
  
  }

  float dist = findPeakDist(correlations, N_SAMP, 10.0);  // Izračun povprečne razdalje med vrhi
  float rate = 1/(dist*MEAS_PERIOD);                      // frekvenca[korak/s] = 1 / perioda[s/korak] ; perioda[s/korak] = perioda[smp/korak]*perioda[s/smp] 
  
  // Preveri če je veljavna frekvenca
  if(!std::isfinite(rate))
  {
    rate = 0.0;
  }

#if PRINT_CORR
  Serial.printf("Rate:%.2f\n\n", rate);
#endif

  return rate;
}

/*

signal     signal, kateremu izračunamo povprečno razdaljo med vrhi
length     dolžina signala
threshold  minimalna vrednost, ki se lahko upošteva kot vrh 


\  /\
 \/  \
      \  /\
       \/  \
    |-----|
*/
float findPeakDist(float* signal, int length, float threshold)
{
  int lastPeak = -1;  // Lokacija zadnjega vrha
  float avg = 0.0;    // Povprečna razdalja med vrhi
  int n = 0;          // Število vrhov
  int state = 0;      // Trenutno stanje (0 - nismo v vrhu, 1 - smo v vrhu)
  float max = 0;      // Največja vrednost trenutnega vrha
  int maxloc = 0;     // Lokacija največje vrednosti trenutnega vrha

  // Zanka čez cel signal
  for(int i = 0; i < length; i++) {
    // Signal nad minimalno mejo
    if(signal[i] > threshold) {
      // Nov vrh ali nova največja vrednost
      if(!state || signal[i] > max) {
        state = 1;
        max = signal[i];
        maxloc = i;
      }
    // Konec vrha
    } else if(state){
      state = 0;
      // Po potrebi prištej razdaljo do zadnjega vrha
      if(lastPeak != -1) {
        avg += maxloc - lastPeak;
      }
      n++;
      lastPeak = maxloc;
    }
  }
  // Upoštevaj tudi zadnjo razdaljo
  if(state && lastPeak != -1)
  {
    avg += maxloc - lastPeak;
    n++;
  }
  avg /= n - 1;

#if PRINT_CORR
  Serial.printf("\nNpeaks:%d, Dist: %.2f\n", n, avg);
#endif

  return avg;
}

void setup() {
  // Serijska komunikacija
  Serial.begin(115200);
  // I2C
  Wire.begin(12,14);
  // SDA - 12 pin
  // SCL - 14 pin
  Wire.setClock(100000);
  // Podesavanje senzorja 
  // https://github.com/bolderflight/mpu9250/blob/main/src/mpu9250.cpp
  MPU9250_init();
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(NAZIV, GESLO);
  // pred nadaljevanjem moramo čakati, dokler WiFi.status() ne vrne WL_CONNECTED
  while (WiFi.status() != WL_CONNECTED){
        delay(500);
        Serial.print(".");
  }
  Serial.printf("WiFi connected\n");

  // Branje senzorja
  readSensor.attach(MEAS_PERIOD, readAcc);
}

// Lastnosti uporabnika
#define HEIGHT 1.8
#define WEIGHT 70

// Akumulirane vrednosti
int steps = 0;
float distance = 0;
float calories = 0;

// Vrednosti za intervale pošiljanja
unsigned long lastConnectionTime = 0;
unsigned long lastUpdateTime = 0;
const unsigned long postingInterval = 15 * 1000; // 15s
const unsigned long updateInterval = 1 * 1000; // 1s

// Buffer za grajenje JSON
char jsonBuffer[4000] = "[";
// updateJson in httpRequest vzeti iz uradne strani thingspeak (zakaj to ni del knjižnice ?)
// https://www.mathworks.com/help/thingspeak/continuously-collect-data-and-bulk-update-a-thingspeak-channel-using-an-arduino-mkr1000-board-or-an-esp8266-board.html
void updateJson(float rate, float stride, float speed){
  /* JSON format for updates paramter in the API
   *  This examples uses the relative timestamp as it uses the "delta_t". You can also provide the absolute timestamp using the "created_at" parameter
   *  instead of "delta_t".
   *   "[{\"delta_t\":0,\"field1\":-70},{\"delta_t\":3,\"field1\":-66}]"
   */
  // Format the jsonBuffer as noted above
  strcat(jsonBuffer,"{\"delta_t\":");
  unsigned long deltaT = (millis() - lastUpdateTime)/1000;
  char temp[10];
  sprintf(temp, "%ld", deltaT);
  strcat(jsonBuffer,temp);
  strcat(jsonBuffer,",");
  strcat(jsonBuffer, "\"field1\":");
  sprintf(temp, "%d,", steps);
  strcat(jsonBuffer,temp);
  strcat(jsonBuffer, "\"field2\":");
  sprintf(temp, "%.2f,", distance);
  strcat(jsonBuffer,temp);
  strcat(jsonBuffer, "\"field3\":");
  sprintf(temp, "%.2f,", calories);
  strcat(jsonBuffer,temp);
  strcat(jsonBuffer, "\"field4\":");
  sprintf(temp, "%.2f,", rate);
  strcat(jsonBuffer,temp);
  strcat(jsonBuffer, "\"field5\":");
  sprintf(temp, "%.2f", speed);
  strcat(jsonBuffer,temp);
  strcat(jsonBuffer,"},");
  // Če je čas od zadnje objave 15 seconds, posodobi ThingSpeak kanal
  if (millis() - lastConnectionTime >=  postingInterval) {
    Serial.printf("heap size: %u\n", ESP.getFreeHeap());
    size_t len = strlen(jsonBuffer);
    jsonBuffer[len-1] = ']';
    httpRequest();
  }
  lastUpdateTime = millis(); // Nastavi zadnji čas posodobitve
}

// Buffer za podatke HTTP zahtevka
char data[5000];
// Funkcija ki objavi JSON zahtevek na ThingSpeak API
void httpRequest() {
  /* JSON format for data buffer in the API
   *  This examples uses the relative timestamp as it uses the "delta_t". You can also provide the absolute timestamp using the "created_at" parameter
   *  instead of "delta_t".
   *   "{\"write_api_key\":\"YOUR-CHANNEL-WRITEAPIKEY\",\"updates\":[{\"delta_t\":0,\"field1\":-60},{\"delta_t\":15,\"field1\":200},{\"delta_t\":15,\"field1\":-66}]
   */
  // Format the data buffer as noted above
  sprintf(data, "{\"write_api_key\":\"%s\",\"updates\":", TS_WRITE_API);
  strcat(data,jsonBuffer);
  strcat(data,"}");
  // Close any connection before sending a new request
  client.stop();
  String data_length = String(strlen(data)+1); //Compute the data buffer length
  Serial.println(data);
  // POST data to ThingSpeak
  if (client.connect("api.thingspeak.com", 80)) {
    client.printf("POST /channels/%ld/bulk_update.json HTTP/1.1\n", CHANNEL);
    client.println("Host: api.thingspeak.com");
    client.println("User-Agent: mw.doc.bulk-update (Arduino ESP8266)");
    client.println("Connection: close");
    client.println("Content-Type: application/json");
    client.println("Content-Length: "+data_length);
    client.println();
    client.println(data);
  }
  else {
    Serial.println("Failure: Failed to connect to ThingSpeak");
  }
  delay(250); //Wait to receive the response
  client.parseFloat();
  String resp = String(client.parseInt());
  Serial.println("Response code:"+resp); // Print the response code. 202 indicates that the server has accepted the response
  jsonBuffer[0] = '['; //Reinitialize the jsonBuffer for next batch of data
  jsonBuffer[1] = '\0';
  lastConnectionTime = millis(); //Update the last conenction time
}

void loop() {
  // Preverjanje za interval posodabljanja
  if((millis() - lastUpdateTime) >= updateInterval)
  {
    // Izračunana frekvenca gibanja
    float rate = calculateSteps();

    // Ocena dolžine koraka
    float height = HEIGHT;
    float stride;
    if(rate < 1) stride = height/5;
    else if(rate < 1.5) stride = height/4;
    else if(rate < 2) stride = height/3;
    else if(rate < 2.5) stride = height/2;
    else if(rate < 3) stride = height/1.2;
    else if(rate < 4) stride = height;
    else stride = 1.2*height;

    // Ocena hitrosti hoje
    float speed = rate*stride;

#if PRINT_CORR
    Serial.printf("Speed: %.2f\n", speed);
#endif

    // Akumuliranje ocenjenih vrednosti
    distance += speed;
    steps += rate;
    float weight = WEIGHT;
    if(rate == 0.0)
      calories += weight/3600;
    else
      calories += speed*weight/800;

    // Posodobitev JSON bufferja
    updateJson(rate, stride, speed);
  }
}